@@ -18,9 +18,10 @@ from nose.tools import assert_equal
 
 import botocore.session
 from botocore.config import Config
-from botocore.compat import urlsplit
-from botocore.compat import parse_qs
+from botocore.compat import datetime, urlsplit, parse_qs
 from botocore.exceptions import ParamValidationError, ClientError
+from botocore.exceptions import InvalidS3UsEast1RegionalEndpointConfigError
+from botocore.parsers import ResponseParserError
 from botocore import UNSIGNED
 
 
@@ -269,6 +270,201 @@ class TestS3ClientConfigResolution(BaseS3ClientConfigurationTest):
             }
         )
 
+    def test_use_arn_region_is_case_insensitive(self):
+        self.environ['AWS_S3_USE_ARN_REGION'] = 'True'
+        client = self.create_s3_client()
+        self.assertEqual(
+            client.meta.config.s3,
+            {
+                'use_arn_region': True,
+            }
+        )
+
+
+    def test_us_east_1_regional_env_var(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        client = self.create_s3_client()
+        self.assertEqual(
+            client.meta.config.s3,
+            {
+                'us_east_1_regional_endpoint': 'regional',
+            }
+        )
+
+    def test_us_east_1_regional_config_var(self):
+        with temporary_file('w') as f:
+            self.set_config_file(
+                f,
+                '[default]\n'
+                's3_us_east_1_regional_endpoint = regional'
+            )
+            client = self.create_s3_client()
+            self.assertEqual(
+                client.meta.config.s3,
+                {
+                    'us_east_1_regional_endpoint': 'regional',
+                }
+            )
+
+    def test_us_east_1_regional_nested_config_var(self):
+        with temporary_file('w') as f:
+            self.set_config_file(
+                f,
+                '[default]\n'
+                's3 = \n'
+                '    us_east_1_regional_endpoint = regional'
+            )
+            client = self.create_s3_client()
+            self.assertEqual(
+                client.meta.config.s3,
+                {
+                    'us_east_1_regional_endpoint': 'regional',
+                }
+            )
+
+    def test_us_east_1_regional_env_var_overrides_config_var(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        with temporary_file('w') as f:
+            self.set_config_file(
+                f,
+                '[default]\n'
+                's3 = \n'
+                '    us_east_1_regional_endpoint = legacy'
+            )
+            client = self.create_s3_client()
+        self.assertEqual(
+            client.meta.config.s3,
+            {
+                'us_east_1_regional_endpoint': 'regional',
+            }
+        )
+
+    def test_client_config_us_east_1_regional_overrides_env_var(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        client = self.create_s3_client(
+            config=Config(
+                s3={'us_east_1_regional_endpoint': 'legacy'}
+            )
+        )
+        self.assertEqual(
+            client.meta.config.s3,
+            {
+                'us_east_1_regional_endpoint': 'legacy',
+            }
+        )
+
+    def test_client_config_us_east_1_regional_overrides_config_var(self):
+        with temporary_file('w') as f:
+            self.set_config_file(
+                f,
+                '[default]\n'
+                's3 = \n'
+                '    us_east_1_regional_endpoint = legacy'
+            )
+            client = self.create_s3_client(
+                config=Config(
+                    s3={'us_east_1_regional_endpoint': 'regional'}
+                )
+            )
+        self.assertEqual(
+            client.meta.config.s3,
+            {
+                'us_east_1_regional_endpoint': 'regional',
+            }
+        )
+
+    def test_client_validates_us_east_1_regional(self):
+        with self.assertRaises(InvalidS3UsEast1RegionalEndpointConfigError):
+            self.create_s3_client(
+                config=Config(
+                    s3={'us_east_1_regional_endpoint': 'not-valid'}
+                )
+            )
+
+    def test_client_region_defaults_to_us_east_1(self):
+        client = self.create_s3_client(region_name=None)
+        self.assertEqual(client.meta.region_name, 'us-east-1')
+
+    def test_client_region_remains_us_east_1(self):
+        client = self.create_s3_client(region_name='us-east-1')
+        self.assertEqual(client.meta.region_name, 'us-east-1')
+
+    def test_client_region_remains_aws_global(self):
+        client = self.create_s3_client(region_name='aws-global')
+        self.assertEqual(client.meta.region_name, 'aws-global')
+
+    def test_client_region_defaults_to_aws_global_for_regional(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        client = self.create_s3_client(region_name=None)
+        self.assertEqual(client.meta.region_name, 'aws-global')
+
+    def test_client_region_remains_us_east_1_for_regional(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        client = self.create_s3_client(region_name='us-east-1')
+        self.assertEqual(client.meta.region_name, 'us-east-1')
+
+    def test_client_region_remains_aws_global_for_regional(self):
+        self.environ['AWS_S3_US_EAST_1_REGIONAL_ENDPOINT'] = 'regional'
+        client = self.create_s3_client(region_name='aws-global')
+        self.assertEqual(client.meta.region_name, 'aws-global')
+
+
+class TestS3Copy(BaseS3OperationTest):
+
+    def create_s3_client(self, **kwargs):
+        client_kwargs = {
+            'region_name': self.region
+        }
+        client_kwargs.update(kwargs)
+        return self.session.create_client('s3', **client_kwargs)
+
+    def create_stubbed_s3_client(self, **kwargs):
+        client = self.create_s3_client(**kwargs)
+        http_stubber = ClientHTTPStubber(client)
+        http_stubber.start()
+        return client, http_stubber
+
+    def test_s3_copy_object_with_empty_response(self):
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-east-1'
+        )
+
+        empty_body = b''
+        complete_body = (
+            b'<?xml version="1.0" encoding="UTF-8"?>\n\n'
+            b'<CopyObjectResult '
+            b'xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
+            b'<LastModified>2020-04-21T21:03:31.000Z</LastModified>'
+            b'<ETag>&quot;s0mEcH3cK5uM&quot;</ETag></CopyObjectResult>'
+        )
+
+        self.http_stubber.add_response(status=200, body=empty_body)
+        self.http_stubber.add_response(status=200, body=complete_body)
+        response = self.client.copy_object(
+            Bucket='bucket',
+            CopySource='other-bucket/test.txt',
+            Key='test.txt',
+        )
+
+        # Validate we retried and got second body
+        self.assertEquals(len(self.http_stubber.requests), 2)
+        self.assertEquals(response['ResponseMetadata']['HTTPStatusCode'], 200)
+        self.assertTrue('CopyObjectResult' in response)
+
+    def test_s3_copy_object_with_incomplete_response(self):
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            region_name='us-east-1'
+        )
+
+        incomplete_body = b'<?xml version="1.0" encoding="UTF-8"?>\n\n\n'
+        self.http_stubber.add_response(status=200, body=incomplete_body)
+        with self.assertRaises(ResponseParserError):
+            self.client.copy_object(
+                Bucket='bucket',
+                CopySource='other-bucket/test.txt',
+                Key='test.txt',
+            )
+
 
 class TestAccesspointArn(BaseS3ClientConfigurationTest):
     _V4_AUTH_REGEX = re.compile(
@@ -771,6 +967,15 @@ class TestRegionRedirect(BaseS3OperationTest):
 
 
 class TestGeneratePresigned(BaseS3OperationTest):
+    def assert_is_v2_presigned_url(self, url):
+        qs_components = parse_qs(urlsplit(url).query)
+        # Assert that it looks like a v2 presigned url by asserting it does
+        # not have a couple of the v4 qs components and assert that it has the
+        # v2 Signature component.
+        self.assertNotIn('X-Amz-Credential', qs_components)
+        self.assertNotIn('X-Amz-Algorithm', qs_components)
+        self.assertIn('Signature', qs_components)
+
     def test_generate_unauthed_url(self):
         config = Config(signature_version=botocore.UNSIGNED)
         client = self.session.create_client('s3', self.region, config=config)
@@ -868,6 +1073,39 @@ class TestGeneratePresigned(BaseS3OperationTest):
         }
         self.assertEqual(parts, expected)
 
+    def test_presign_uses_v2_for_aws_global(self):
+        client = self.session.create_client('s3', 'aws-global')
+        url = client.generate_presigned_url(
+            'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
+        self.assert_is_v2_presigned_url(url)
+
+    def test_presign_uses_v2_for_default_region_with_us_east_1_regional(self):
+        config = Config(s3={'us_east_1_regional_endpoint': 'regional'})
+        client = self.session.create_client('s3', config=config)
+        url = client.generate_presigned_url(
+            'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
+        self.assert_is_v2_presigned_url(url)
+
+    def test_presign_uses_v2_for_aws_global_with_us_east_1_regional(self):
+        config = Config(s3={'us_east_1_regional_endpoint': 'regional'})
+        client = self.session.create_client('s3', 'aws-global', config=config)
+        url = client.generate_presigned_url(
+            'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
+        self.assert_is_v2_presigned_url(url)
+
+    def test_presign_uses_v2_for_us_east_1(self):
+        client = self.session.create_client('s3', 'us-east-1')
+        url = client.generate_presigned_url(
+            'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
+        self.assert_is_v2_presigned_url(url)
+
+    def test_presign_uses_v2_for_us_east_1_with_us_east_1_regional(self):
+        config = Config(s3={'us_east_1_regional_endpoint': 'regional'})
+        client = self.session.create_client('s3', 'us-east-1', config=config)
+        url = client.generate_presigned_url(
+            'get_object', {'Bucket': 'mybucket', 'Key': 'mykey'})
+        self.assert_is_v2_presigned_url(url)
+
 
 def test_correct_url_used_for_s3():
     # Test that given various sets of config options and bucket names,
@@ -1095,6 +1333,20 @@ def test_correct_url_used_for_s3():
         s3_config=use_dualstack, signature_version='s3',
         # Still default to virtual hosted when possible on sigv2.
         expected_url='https://bucket.s3.dualstack.us-east-1.amazonaws.com/key')
+    yield t.case(
+        region=None, bucket='bucket', key='key',
+        s3_config=use_dualstack,
+        # Uses us-east-1 for no region set.
+        expected_url='https://bucket.s3.dualstack.us-east-1.amazonaws.com/key')
+    yield t.case(
+        region='aws-global', bucket='bucket', key='key',
+        s3_config=use_dualstack,
+        # Pseudo-regions should not have any special resolving logic even when
+        # the endpoint won't work as we do not have the metadata to know that
+        # a region does not support dualstack. So just format it based on the
+        # region name.
+        expected_url=(
+            'https://bucket.s3.dualstack.aws-global.amazonaws.com/key'))
     yield t.case(
         region='us-west-2', bucket='bucket', key='key',
         s3_config=use_dualstack, signature_version='s3',
@@ -1410,6 +1662,67 @@ def test_correct_url_used_for_s3():
         )
     )
 
+    # Use us-east-1 regional endpoint cases: regional
+    us_east_1_regional_endpoint = {
+        'us_east_1_regional_endpoint': 'regional'
+    }
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint,
+        expected_url=(
+            'https://bucket.s3.us-east-1.amazonaws.com/key'))
+    yield t.case(
+        region='us-west-2', bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint,
+        expected_url=(
+            'https://bucket.s3.us-west-2.amazonaws.com/key'))
+    yield t.case(
+        region=None, bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint,
+        expected_url=(
+            'https://bucket.s3.amazonaws.com/key'))
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config={
+            'us_east_1_regional_endpoint': 'regional',
+            'use_dualstack_endpoint': True,
+        },
+        expected_url=(
+            'https://bucket.s3.dualstack.us-east-1.amazonaws.com/key'))
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config={
+            'us_east_1_regional_endpoint': 'regional',
+            'use_accelerate_endpoint': True,
+        },
+        expected_url=(
+            'https://bucket.s3-accelerate.amazonaws.com/key'))
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config={
+            'us_east_1_regional_endpoint': 'regional',
+            'use_accelerate_endpoint': True,
+            'use_dualstack_endpoint': True,
+        },
+        expected_url=(
+            'https://bucket.s3-accelerate.dualstack.amazonaws.com/key'))
+
+    # Use us-east-1 regional endpoint cases: legacy
+    us_east_1_regional_endpoint_legacy = {
+        'us_east_1_regional_endpoint': 'legacy'
+    }
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint_legacy,
+        expected_url=(
+            'https://bucket.s3.amazonaws.com/key'))
+
+    yield t.case(
+        region=None, bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint_legacy,
+        expected_url=(
+            'https://bucket.s3.amazonaws.com/key'))
+
 
 class S3AddressingCases(object):
     def __init__(self, verify_function):
@@ -1577,6 +1890,21 @@ def test_addressing_for_presigned_urls():
         )
     )
 
+    # Use us-east-1 regional endpoint configuration cases
+    us_east_1_regional_endpoint = {
+        'us_east_1_regional_endpoint': 'regional'
+    }
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint, signature_version='s3',
+        expected_url=(
+            'https://bucket.s3.us-east-1.amazonaws.com/key'))
+    yield t.case(
+        region='us-east-1', bucket='bucket', key='key',
+        s3_config=us_east_1_regional_endpoint, signature_version='s3v4',
+        expected_url=(
+            'https://bucket.s3.us-east-1.amazonaws.com/key'))
+
 
 def _verify_presigned_url_addressing(region, bucket, key, s3_config,
                                      is_secure=True,
